# -*- coding: utf8 -*-
import re
import string
import imLib.Category as Category
import imLib.config as config

class ListParser:
	listMarkPat = r'\*|(?:[a-zA-Z0-9]\.)'
	def __init__(self):
		listitemPat = r'^ +(?:%(listMarkPat)s) ' % {'listMarkPat':self.listMarkPat}
		self.rules = re.compile(listitemPat, re.DOTALL|re.MULTILINE)
	def feed(self, rawText):
		self.rawText = rawText
	def parse(self):
		start = 0
		self.result = []
		self.indentList = []
		self.totalIndentCount = 0
		i =1
		for match in self.rules.finditer(self.rawText):
			i = i + 1
			end = match.start()
			item = self.rawText[start:end]
			start = end
			if not item: # first
				continue
			lastIndent = self.appendToResult(item)
		item = self.rawText[start:]
		self.appendToResult(item)
		return self.result

	def appendToResult(self, item):
		curIndent, listType, contents = self.split(item)
		if self.indentList:
			lastIndent = self.indentList[-1]
		else:
			lastIndent = 0
		indentCount = 0
		if curIndent > lastIndent:
			indentCount = 1
			self.indentList.append(curIndent)
		elif curIndent < lastIndent:
			lastIndent = self.indentList.pop()
			while self.indentList and curIndent < lastIndent:
				lastIndent = self.indentList.pop()
				indentCount = indentCount - 1
			if curIndent >= lastIndent:
				self.indentList.append(lastIndent)
		self.result.append((indentCount, listType, contents))
		return curIndent


	def split(self, item):
		curIndent = 0
		while item[curIndent] == ' ':
			curIndent = curIndent + 1
		listType = self.getListType(item[curIndent])
		start = listType and curIndent+3 or curIndent+2
		return curIndent, listType, item[start:]
	def getListType(self, c): # 두번째 return은 strict로 할때를 위한..
		if c == 'i':
			return 'i'
			return 'lower-roman'
		elif c == 'I':
			return 'I'
			return 'upper-roman'
		elif c in string.lowercase:
			return 'a'
			return 'lower-alpha'
		elif c in string.uppercase:
			return 'A'
			return 'upper-alpha'
		elif c in string.digits:
			return '1'
			return 'decimal'
		return None # * => None


class TableParser:
	def __init__(self, sep='||'):
		self.sep = sep
	def feed(self, rawText):
		self.rawText = rawText
	def parse(self):
		self.rowList = []
		for line in self.rawText.strip().splitlines():
			colList = line.strip().split(self.sep)[1:-1]
			L = []
			for col in colList:
				align, col = self.getAlign(col)
				if align:
					L.append((align, col))
				else:
					L.append(col)
			self.rowList.append(L)
		return self.rowList
	def getAlign(self, col):			
		if len(col) < 2:
			return '', col
		if col[0:2] in ['><', '<>']:
			align = 'center'
			col = col[2:]
		elif col[0] == '<':
			align = 'left'
			col = col[1:]
		elif col[0] == '>':
			align = 'right'
			col = col[1:]
		else:
			align = ''
		return align, col

class IndentParser:
	def __init__(self):
		pass
	def feed(self, rawText):
		self.rawText = rawText
	def parse(self):
		lines = self.rawText.splitlines() 
		lastIndent = 0
		lastIndentList = []
		self.result = []
		totalIndentCount = 0
		for i in range(len(lines)):
			line = lines[i]
			curIndent , contents = self.split(line)
			if curIndent == 0 or curIndent == lastIndent:
				self.insertToPrevContents(contents)  # curIndent == 0 -> contents == \r, \n
				continue
			if curIndent > lastIndent:
				indentCount = 1
				totalIndentCount = totalIndentCount + indentCount
			elif curIndent < lastIndent:	
				indentCount, lastIndent = self.dedentCounter(curIndent, lastIndent, lastIndentList)
				# FIXME vvvvvvvvvv refactoring
				if totalIndentCount < abs(indentCount):
					indentCount = 0 # 원래는 계산을 해야겠지만 어차피 이에 해당하는 경우는 totalIndentCount = 0이고 indentCount는 -1인 경우뿐이 없다.
					totalIndentCount = 0
				else:
					totalIndentCount = totalIndentCount + indentCount
				# FIXME ^^^^^^^^^^^^^^^^

			lastIndentList.append(lastIndent)
			lastIndent = curIndent
			self.result.append((indentCount, contents+'\n'))
		return self.result
	def insertToPrevContents(self, contents):
		indentCount, prevContents = self.result[-1]
		self.result[-1] = (indentCount, prevContents + contents+'\n')
	def dedentCounter(self, curIndent, lastIndent, lastIndentList):
		indentCount = 0
		while curIndent < lastIndent:
			indentCount = indentCount - 1
			lastIndent = lastIndentList.pop()
		return indentCount, lastIndent

	def split(self, line):
		indentLevel = 0
		for c in line:
			if c == ' ': indentLevel = indentLevel + 1
			else: break
		return indentLevel, line[indentLevel:]

markUpMap = {
	'heading':('strike', 'underline', 'internalLink', 'camelCase'),
	'pre':(),
	'tt':(),

	'blockquote':('tt', 'strong', 'em', 'strongEm', 'strike', 'underline', 'superscript', 'subscript', 'internalLink', 'camelCase', 'externalLink', 'smiley', 'horizon', 'br', 'indent', 'list', 'table'), 
	'inlinequote':('tt', 'strong', 'em', 'strongEm', 'strike', 'underline', 'superscript', 'subscript', 'internalLink', 'camelCase', 'externalLink', 'smiley'),
	'strong':('tt', 'inlinequote', 'strike', 'underline', 'superscript', 'subscript', 'internalLink', 'camelCase', 'externalLink', 'smiley'), 
	'strongEm':('tt', 'inlinequote', 'strike', 'underline', 'superscript', 'subscript', 'internalLink', 'camelCase', 'externalLink', 'smiley'), 
	'em':('tt', 'inlinequote', 'strike', 'underline', 'superscript', 'subscript', 'internalLink', 'camelCase', 'externalLink', 'smiley'), 
	'strike':('tt', 'inlinequote', 'strongEm', 'strong', 'em', 'underline', 'superscript', 'subscript', 'internalLink', 'camelCase', 'externalLink', 'smiley'),
	'underline':('tt', 'inlinequote', 'strongEm', 'strong', 'em', 'strike', 'superscript', 'subscript', 'internalLink', 'camelCase', 'externalLink', 'smiley'),
	'superscript':('tt', 'inlinequote', 'strongEm', 'strong', 'em', 'strike', 'underline', 'subscript', 'internalLink', 'camelCase', 'externalLink', 'smiley'),
	'subscript':('tt', 'inlinequote', 'strongEm', 'strong', 'em', 'strike', 'underline', 'superscript', 'internalLink', 'camelCase', 'externalLink', 'smiley'),

	'table':('inlinequote', 'strong', 'strongEm', 'em', 'strike', 'underline', 'internalLink', 'tt', 'camelCase', 'externalLink', 'macro', 'smiley'),
	'list':('inlinequote', 'indent', 'strong', 'strongEm', 'em', 'strike', 'underline', 'tt', 'internalLink', 'camelCase', 'externalLink', 'macro', 'smiley', 'br'),
	'indent':('inlinequote', 'table', 'tt', 'strong', 'strongEm', 'em', 'strike', 'underline', 'listitem', 'internalLink', 'camelCase', 'externalLink', 'macro', 'br', 'smiley'),

	'macro':(),
	'internalLink':(),
	'camelCase':(),
	'externalLink':(),

	'br':(),
	'horizon':(),
	'smiley':(),
}
class Parser:
	def __init__(self, formatter):
		self.formatter = formatter

		patterns = r"""
(^(?P<heading_marker>={1,6})( +)(?P<heading>[^\n]+?)( +)(?P=heading_marker)$[\n]?)
(^\{{3}$(?P<pre>.+?)^\}{3} *?$\n?)
(\{{3}(?P<tt>.+?)\}{3})
(^"{3}$\n(?P<blockquote>.+?)\n^"{3} *?$\n?)
("{3}(?P<inlinequote>[^"\n].+?)"{3})
('{5}(?P<strongEm>[^' ].+?)'{5})
('{3}(?P<strong>[^' ].+?)'{3})
('{2}(?P<em>[^' ].+?)'{2})
(~{2}(?P<strike>[^~ ].+?)~{2})
(_{2}(?P<underline>[^_ ].+?)_{2})
(\^{2}(?P<superscript>[^\^ ].+?)\^{2})
(,{2}(?P<subscript>[^, ].+?),{2})
(?P<horizon>^-{4,}[\n]{0,1}$)
(?P<table>(^(\|\|[^\n]*?)+\|\|$[\n]?)+)
(?P<list>^ +(%(listMark)s) .+?$\n?(^( +[^\n]*)|\n)*)
(?P<indent>(^ +(?!%(listMark)s)[^\n]+$[\n]*)+)
(\[\[(?P<macro>([^\[\n]+?))\]\])
(\[(?P<internalLink>([^\[]+?))\])
(?P<camelCase>/?(([A-Z][a-z]+){2,}/?)+)
(?P<externalLink>http://\S+)
(?P<br>[\n])
(?P<smiley>%(smiley)s)""" % {

		'listMark' : ListParser.listMarkPat,
		'smiley' : '|'.join(map(re.escape, config.smileyDict.keys())),
		}
		patterns = patterns.strip().replace('\n', '|')
		self.rules = re.compile(patterns, re.DOTALL|re.MULTILINE)

		self.tableParser = TableParser()
		self.listParser = ListParser()

	def feed(self, rawText):
		self.rawText = rawText.replace('\r', '')

	def parse(self):
		self.stackOfopenedMarkUp = []
		self.parsedText = self.rawText
		self.parsedText = self.rules.sub(self.replaceAgent, self.parsedText)
		return self.parsedText

	# replace methods
	def replaceAgent(self, match):
		D = match.groupdict()
		del D['heading_marker']
		type, matchedStr = filter(lambda group: group[1], D.items())[0]
		replFunc = getattr(self, '_repl%s' % type.capitalize())
		if self.stackOfopenedMarkUp and type not in markUpMap[self.stackOfopenedMarkUp[-1]]: #FIXME 매우 급함. 매우 치명적
			matchedStr = match.group()
			return matchedStr[0] + self.rules.sub(self.replaceAgent, matchedStr[1:]) # 임시
			return matchedStr[0]  # FIXME 이거 문제네..A마크업 안에서 B 마크업이 무시되고 C 마크업은 유효할 때, A - B - C 이런순으로 감싸놨으면 C가 유효해야하는데 무효가 되버린다.
		insertTheMatchObjList = ['heading'] #FIXME my name is ...-_-
		self.stackOfopenedMarkUp.append(type)
		if type in insertTheMatchObjList:
			result = replFunc(match)
		else:
			result = replFunc(matchedStr)
		self.stackOfopenedMarkUp.pop()
		return result

	def _replStrong(self, matchedStr):
		return self.formatter.strong(self.rules.sub(self.replaceAgent, matchedStr))
	def _replEm(self, matchedStr):
		return self.formatter.em(self.rules.sub(self.replaceAgent, matchedStr))
	def _replStrongem(self, matchedStr):
		return self.formatter.strongEm(self.rules.sub(self.replaceAgent, matchedStr))
	def _replBlockquote(self, matchedStr):
		return self.formatter.blockquote(self.rules.sub(self.replaceAgent, matchedStr))
	def _replInlinequote(self, matchedStr):
		return self.formatter.q(self.rules.sub(self.replaceAgent, matchedStr))

	def _replStrike(self, matchedStr):
		return self.formatter.strike(self.rules.sub(self.replaceAgent, matchedStr))
	def _replUnderline(self, matchedStr):
		return self.formatter.underline(self.rules.sub(self.replaceAgent, matchedStr))
	def _replSuperscript(self, matchedStr): # FIXME testcode
		return self.formatter.superscript(self.rules.sub(self.replaceAgent, matchedStr))
	def _replSubscript(self, matchedStr): # FIXME testcode..
		return self.formatter.subscript(self.rules.sub(self.replaceAgent, matchedStr))
	def _replHorizon(self, matchedStr): # FIXME
		return self.formatter.horizon(matchedStr)
	def _replPre(self, matchedStr):
		return self.formatter.pre(matchedStr)
	def _replHeading(self, match):
		title = match.groupdict()['heading']
		level = len(match.groupdict()['heading_marker'])
		return self.formatter.heading(title, level)

	def _replTable(self, matchedStr):
		self.tableParser.feed(matchedStr)
		rowList = self.tableParser.parse()
		# FIXME rowList도 '''등은 해줘야지.
		return self.formatter.table(rowList)

	def _replList(self, matchedStr): # FIXME replList, replIndent 너무 비슷함
		self.listParser.feed(matchedStr)
		itemList = self.listParser.parse()
		L = []
		for indentCount, listType, contents in itemList:
			parsedContents = self.formatter.listitem(self.rules.sub(self.replaceAgent, contents))
			L.append((indentCount, listType, parsedContents))
		return self.formatter.List(L)

	def _replIndent(self, matchedStr): # FIXME
		indentParser = IndentParser()
		indentParser.feed(matchedStr)
		L = []
		parsedList = indentParser.parse()
		newParsedList = []
		for indentLevel, parsedText in parsedList:
			parsedText = self.rules.sub(self.replaceAgent, parsedText)
			newParsedList.append((indentLevel, parsedText))
		return self.formatter.indent(newParsedList)

	def _replListitem(self, match):  # FIXME 고쳐라.
		type, matchedStr = filter(lambda group: group[1], match.groupdict().items())[0]
		stackLength = len(self.stackOfopenedMarkUp)
		if stackLength < 2 or (stackLength >= 2 and self.stackOfopenedMarkUp[-2] not in  ['indent', 'list']): 
			return self.rules.sub(self.replaceAgent, match.group()) 
		return self.formatter.listitem(self.rules.sub(self.replaceAgent, matchedStr))
		#return self.formatter.listitem(self.rules.sub(self.replaceAgent, matchedStr.lstrip()))

	def _replTt(self, matchedStr):
		return self.formatter.tt(matchedStr)

	def _replInternallink(self, matchedStr):
		if matchedStr.startswith('http://'):  #FIXME 지져분하다. 정리해라. 
			firstSpace = matchedStr.find(' ')
			if firstSpace > 0:
				url = matchedStr[:firstSpace]
				linkName = matchedStr[firstSpace+1:]
				return self.formatter.externalLink(url, linkName)
			return self.formatter.externalLink(matchedStr)

		L = matchedStr.split(Category.sep)
		pageName = L[-1]
		categoryPath = Category.sep.join(L[:-1])
		isAbsolutePath = matchedStr and not matchedStr.startswith(Category.sep)
		return self.formatter.internalLink(categoryPath, pageName, isAbsolutePath)
	def _replCamelcase(self, matchedStr):
		return self._replInternallink(matchedStr)
	def _replExternallink(self, matchedStr):
		return self.formatter.externalLink(matchedStr)

	def _replMacro(self, matchedStr):
		return self.formatter.macro(matchedStr)
		return matchedStr
		L = matchedStr.split('(')
		if len(L) > 1:
			macroName = L[0]
			macroArgs = L[-1]
		return matchedStr
	def _replBr(self, matchedStr):
		if 'pre' in self.stackOfopenedMarkUp:
			return matchedStr
		return '<br />'
	def _replSmiley(self, matchedStr):
		return self.formatter.smiley(matchedStr)

class EchoFormatter: # FOR UNITTEST
	def __init__(self):
		self._stack = []
	def echo(self, *args):
		s = ' : '.join([str(arg) for arg in args])
		key = self._stack.pop()
		return '<%s>%s</%s>' % (key, s, key)
	def __getattr__(self, key):
		self._stack.append(key)
		return self.echo
